#include <iostream>
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <map>
#include <string>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <unistd.h>
#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

using namespace std;
#include "misc.hh"

unsigned int page_size;
int page_mask;
int page_blocks;

static void setpagevalues()
{
  page_size=getpagesize();
  page_mask=(~(page_size-1));
  page_blocks=page_size/1024;
}

class FaultGatherer
{
public:
  FaultGatherer()
  {
    struct rusage ru;    
    getrusage(RUSAGE_SELF, &ru);
    d_minor=ru.ru_minflt;
    d_major=ru.ru_majflt;
  }
  ~FaultGatherer()
  {
    struct rusage ru;    
    getrusage(RUSAGE_SELF, &ru);
    s_minor+=ru.ru_minflt-d_minor;
    s_major+=ru.ru_majflt-d_major;
  }
  static void getFaultsZ(long *minor, long *major) 
  {
    *minor=s_minor;
    *major=s_major;
    s_minor=s_major=0;
  }
private:
  static long s_minor, s_major;
  long d_minor, d_major;
};

long FaultGatherer::s_minor;
long FaultGatherer::s_major;

int getLine(string &line)
{
  char *p=readline("vmloader> ");
  if(!p)
    return 0;
  add_history(p);
  line=p;
  free(p);
  return 1;
}

char *arena;
unsigned int mbsize;
bool haveArena, haveMap;
typedef void cmdfunc( const vector<string>& words);

struct function
{
  function()
  {
    ptr=0;
    help=0;
  }
  function(cmdfunc *p, char *h)
  {
    ptr=p;
    help=h;
  }
  cmdfunc *ptr;
  char *help;
};
typedef map<string,function>functions_t;
functions_t functions;

void printrUsage( const vector<string>& words)
{
  long minor, major;
  FaultGatherer::getFaultsZ(&minor, &major);
  cout<<"minor: "<<minor<<", major: "<<major<<endl;
}

int rangeGet(const vector<string>& words, unsigned int* begin, unsigned int* end )
{
  if(!haveMap && ! haveArena) {
    cerr<<"No arena!"<<endl;
    return -1;
  }

  if(words.size()==3) {
    *begin=atoi(words[1].c_str());
    *end=atoi(words[2].c_str());
  }
  if(words.size()==2) {
    *begin=0;
    *end=atoi(words[1].c_str());
  }
  if(*begin>mbsize || *end>mbsize) {
    cerr<<"Tried to touch outside of arena"<<endl;
    return -1;
  }
  return 0;
}

void touch(const vector<string>& words)
{
  unsigned int begin=0, end=mbsize;
  if(rangeGet(words, &begin,&end))
    return;

  if(words[0]=="sweep") 
    cout<<"Sweeping";
  else
    cout<<"Touching";

  cout<<" from mbyte "<<begin<<" to "<<end<<", "<<(end-begin)*1000/4<<" pages";
  cout.flush();

  volatile int dummy;
  {
    FaultGatherer FG;
    if(words[0]=="sweep")
      for(unsigned int n=begin*1000;n<end*1000;n++)
	dummy+=*(arena+n*1024);
    else
      for(unsigned int n=begin*1000;n<end*1000;n++)
	*(arena+n*1024)=n&0xff;
  }

  cout<<". Done"<<endl;
}

void rtouch(const vector<string>& words)
{
  unsigned int begin=0, end=mbsize;
  if(rangeGet(words, &begin, &end))
    return;

  vector<char *>positions;

  for(unsigned int n=begin*1000;n<end*1000;n++)
    positions.push_back(arena+n*1024);
  random_shuffle(positions.begin(),positions.end());
  
  cout<<"Random ";
  if(words[0]=="rsweep") 
    cout<<"sweeping";
  else
    cout<<"touching";

  cout<<" from mbyte "<<begin<<" to "<<end<<", "<<(end-begin)*1000/4<<" pages";
  cout.flush();

  volatile int dummy;
  {
    FaultGatherer FG;
    if(words[0]=="rsweep")
      for(vector<char *>::const_iterator i=positions.begin();i!=positions.end();++i)
	dummy+=*(*i);
    else
      for(vector<char *>::const_iterator i=positions.begin();i!=positions.end();++i)
	*(*i)=(int)(*i)&0xff;
  }

  cout<<". Done"<<endl;
}

void mkfile(const vector<string>& words)
{
  if(words.size()!=3) {
    cerr<<"syntax: "<<words[0]<<" filename megabytes"<<endl;
    return;
  }
  FILE *fp=fopen(words[1].c_str(),"w");
  if(!fp) {
    cerr<<"Failed to open "<<words[1]<<" for writing: "<<strerror(errno)<<endl;
    return;
  }
  char *p=new char[page_size];
  int blocks=atoi(words[2].c_str())*1000/page_blocks;
  cout<<"Writing "<<blocks<<" pages";
  cout.flush();

  int n;
  for(n=0;n<blocks;++n)
    if(fwrite(p,1,page_size,fp)!=page_size) 
      break;

  if(n!=blocks) {
    cerr<<". Only wrote "<<n<<" pages, deleting file"<<endl;
    fclose(fp);
    unlink(words[1].c_str());
    return;
  }
  if(fclose(fp)) {  // NFS dorks
    cerr<<". Error closing file, deleting"<<endl;
    unlink(words[1].c_str());
  }
  else
    cout<<". Done"<<endl;
}

void unmapfile(const vector<string>& words)
{
  if(!haveMap) {
    cerr<<"Nothing to unmap"<<endl;
    return;
  }
  munmap(arena, mbsize*1000*page_size);
  haveMap=0;
}


void mapfile(const vector<string>& words)
{
  if(haveArena) {
    cerr<<"There is an allocted arena, free that first"<<endl;
    return;
  }
  if(haveMap) {
    cerr<<"There is file mapped already, unmap that first"<<endl;
    return;
  }
  if(words.size()!=2) {
    cerr<<"Syntax: "<<words[0]<<" filename"<<endl;
    return;
  }
  int fd=open(words[1].c_str(),O_RDWR);
  if(fd<0) {
    cerr<<"Opening file "<<words[1]<<": "<<strerror(errno)<<endl;
    return;
  }
  struct stat sb;
  if(fstat(fd,&sb)<0) {
    cerr<<"error determining filesize of "<<words[1]<<": "<<strerror(errno)<<endl;
    close(fd);
    return;
  }
  int pages=sb.st_size/page_size;     // 1024000 bytes = 1 megabyte, 256 pages
  int mbytes=pages*page_blocks/1000;  // 1
  if(!mbytes) {
    cerr<<"File less than 1 megabyte, too small for us"<<endl;
    close(fd);
    return;

  }
  cout<<"mmap()ing "<<pages<<" pages ("<<mbytes<<" mbytes)";
  cout.flush();
  arena=(char *)mmap(0,(size_t)mbytes*1000*1024,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
  close(fd);
  if(arena==(void *)-1) {
    cerr<<". Failed: "<<strerror(errno)<<endl;
     return;
  }
  haveMap=true;
  mbsize=mbytes;
  cout<<". Done"<<endl;
  
}

void mincoreStats(const vector<string>& words)
{
  if(!haveMap) {
    cerr<<"No map"<<endl;
    return;
  }
  int numpages=mbsize*1000/page_blocks;
  
  unsigned char *vect=new unsigned char[numpages];
  if(mincore((void *)arena, numpages*page_size,vect)) { // let's hope arena is page aligned 
    cerr<<"mincore returned error: "<<strerror(errno)<<endl;
    return;
  }
  int page=0;
  int incore=0;
  for(page=0;page<numpages;++page)
    if(vect[page]&1)
      incore++;
  cout<<incore<<" pages of "<<numpages<<" available in memory"<<endl;
}

void allocate(const vector<string>& words)
{
  int mbytes;

  if(words[0]=="free") {
    mbytes=0;
  } 
  else {
    if(words.size()!=2) {
      cerr<<words[0]<<" takes 1 parameter"<<endl;
      return;
    }
    mbytes=atoi(words[1].c_str());
  }

  if(haveMap) {
    cerr<<"There is a map in place, unmap it first"<<endl;
    return;
  }

  if(!mbytes) {
    if(!haveArena) {
      cerr<<"There is no arena"<<endl;
      return;
    }
    free(arena);
    mbsize=0;
    cout<<"Freed arena"<<endl;
    haveArena=false;
    return;
  }
  else if(haveArena) {
    char *newarena=(char *)realloc(arena,mbytes*1000*1024);
    if(!newarena) {
      cerr<<"Did not succeed in growing from "<<mbsize<<" to "<<mbytes<<" megabytes, leaving old arena"<<endl;

      return;
    }
    if(newarena!=arena) 
      cerr<<"Warning: VMA was moved"<<endl;

    arena=newarena;
    mbsize=mbytes;
    cout<<"Arena now "<<mbsize<<" megabytes, "<<mbsize*1000/page_blocks<<" pages"<<endl;
    return;
  }
  else {
    if(!(arena=(char *)malloc(mbytes*1000*1024))) {
      cerr<<"Did not succeed in allocating "<<mbytes<<" megabytes, no arena"<<endl;
      return;
    }
    mbsize=mbytes;
    haveArena=true;
    cout<<"Arena now "<<mbsize<<" megabytes, "<<mbsize*1000/page_blocks<<" pages"<<endl;
  }
  
}

void quit(const vector<string>&words)
{
  exit(1);
}

void help(const vector<string>&words)
{
  int printed=0;
  for(functions_t::const_iterator i=functions.begin();i!=functions.end();++i) 
    if(words.size()==1 || i->first.substr(0,words[1].size())==words[1]) {
      cout<<i->first<<"\t\t"<<i->second.help<<endl;
      printed++;
    }
  if(!printed) {
    cerr<<"No command matched"<<endl;
  }
  return;
        
}


void warranty(const vector<string>&words)
{
  cerr<<"None"<<endl;
}


void initFuncs()
{
  functions["faults"]=function(&printrUsage,"Print faults");
  functions["alloc"]=function(&allocate,"Allocate or reallocate memory");
  functions["free"]=function(allocate,"Free the arena");
  functions["touch"]=function(&touch,"Touch pages. Accepts a range");
  functions["sweep"]=function(&touch,"Read pages. Accepts a range");
  functions["rtouch"]=function(&rtouch,"Touch pages in random order. Accepts a range");
  functions["rsweep"]=function(&rtouch,"Read pages in random order. Accepts a range");
  functions["quit"]=function(&quit,"Exit vmloader");
  functions["help"]=function(&help,"Prints this useful message");
  functions["mkfile"]=function(&mkfile,"Makes an n megabyte file");
  functions["mmap"]=function(&mapfile,"Map a file in memory");
  functions["map"]=function(&mapfile,"Map a file in memory (alias)");
  functions["unmap"]=function(&unmapfile,"Unmap a file in memory");
  functions["mincore"]=function(&mincoreStats,"How much of the map is in memory");
  functions["warranty"]=function(&warranty,"If it breaks, you get to keep both parts");
}



int getAbbrev(string &command)
{
  if(functions.count(command))
    return 1;

  int matches=0;
  string potcommand;
  for(functions_t::const_iterator i=functions.begin();i!=functions.end();++i) {
    if(i->first.substr(0,command.size())==command) {
      matches++;
      potcommand=i->first;
    }
  }
  if(matches==1)
    command=potcommand;

  return matches;

}

void welcome()
{
  cout<<"vmloader version "VERSION", Copyright (C) 2002  bert hubert"<<endl;
  cout<<"vmloader comes with ABSOLUTELY NO WARRANTY; for details type `warranty'."<<endl;
  cout<<"This is free software, and you are welcome to redistribute it"<<endl;
  cout<<"under certain conditions."<<endl<<endl;
}


int main(int argc, char **argv)
{
  welcome();
  
  string line;
  vector<string> words;
  string command;

  
  setpagevalues();
  initFuncs();



  int numCommands;
  while(getLine(line)) {
    words.clear();
    stringtok(words,line);
    if(words.empty())
      continue;
    
    numCommands=getAbbrev(words[0]);
    if(numCommands==1) {
      (*functions[words[0]].ptr)(words);      
    }
    else if(numCommands>1)
      cerr<<"Ambiguous command"<<endl;
    else
      cerr<<"No such command, '"<<words[0]<<"'"<<endl;
     
  }
}
